package Sophos::Reports::Report::QueueStatus;
use strict;
use warnings;
use base 'Sophos::Reports::Report::PMX';

use PureMessage::Version qw($PREFIX);
use PureMessage::Quarantine;
use LWP::UserAgent;

sub new {
    my ($class, %opts) = @_;
    my $self = $class->SUPER::new(%opts);

    $self->{type} ||= 'table';
    $self->{views} = {
       table => 1,
    };
    $self->{queues} =
        [map { ucfirst }
             qw(default active corrupt deferred hold incoming maildrop)];
    $self->{server} ||= '';
    $self->{queue} ||= '';
    $self->{max_subdomains} ||= 20;

    if ($self->{queue} and !grep {$self->{queue} eq $_} @{$self->{queues}}) {
        die "Invalid queue: $self->{queue}\n"; # XXX localize?
    }

    if ($self->{server} and !grep {$self->{server} eq $_} @{ $self->servers }) {
        die "Invalid server: $self->{server}\n"; # XXX localize?
    }

    # super will setup lang
    $self->{title} = "/groups/images/$self->{lang}/ttl_mt_queueStatus.png",
    return $self;
}

sub _init_from_uri {
    my ($self, $uri) = @_;

    my ($parameters, $output) = $uri =~ m#^([^.?]+)(?:\.([^/]*))?#
	or die "Invalid parameters";

    die "No parameters" unless $parameters;

    $self->{output} = $output || 'html';

    my @split = split(m{/}, $parameters);

    $self->{type} = shift @split || die "No report type specified";
    $self->{queue} = shift @split || '';

    die "Too many parameters" if @split;
}

sub has_direction { 0 }

sub queues { $_[0]->{queues} }
sub server { $_[0]->{server} }
sub field_names { $_[0]->{field_names} }

sub queue { 
    my $self = shift;
    return $self->{queue} || 'Default';
}

sub servers {
    my $self = shift;
    unless ($self->{servers}) {
        my $pmdb = PureMessage::Quarantine->new;
        $self->{servers} = $pmdb->active_edge_servers;
    }
    return $self->{servers};
}

sub export_src {
    my $self = shift;
    (my $class = ref($self)) =~ s#.+::(.+)#$1#;
    my $queue = $self->queue;
    my $url = "$class/$self->{type}/$queue";
    $url .= "/$self->{server}" if $self->{server};
    return "$self->{uri_prefix}/$url.csv";
}

sub data {
    my ($self) = @_;
    my $data = $self->_data;

    if (@$data > 1) {
        my @totals = ('TOTAL'); # XXX - localize TOTAL?
        for my $host (@$data) {
            for my $i (1 .. $#$host) {
                $totals[$i] += $host->[$i];
            }
        }
        unshift @$data, \@totals;;
    }
    return $data;
}

sub csv {
    my $self = shift;
    my $data = $self->_data;
    unshift @$data, $self->{field_names};
    my $csv = '';
    for my $d (@$data) {
        $csv .= join(",", @$d) . "\n";
    }
    return $csv;
}

sub _data {
    my $self = shift;
    my $servers = $self->{server} ? [ $self->{server} ]
                                  : $self->servers;
    my %qshape;
    for my $server (@$servers) {
        my $data;
        eval {
            $data = $self->_fetch_qshape($server);
        };
        if ($@) {
            warn $@;
            next;
        }
        for my $line (split "\n", $data) {
            $line =~ s/^\s+(.+?)\s*$/$1/; # trim whitespace
            my @tokens = split /\s+/, $line;

            if ($tokens[0] eq 'T') {
                shift @tokens;
                $self->{field_names} = [ "HOST", "TOTAL", @tokens ];
                next;
            }

            my $host = shift @tokens;
            next if $host eq 'TOTAL';
            if ($qshape{$host}) {
                for my $i (0 .. $#tokens) {
                    $qshape{$host}->[$i] += $tokens[$i];
                }
            }
            else {
                $qshape{$host} = \@tokens;
            }
        }
    }

    # sort - most queued messages first (per-host), then hostname
    my @data = map { [ $_, @{ $qshape{$_} } ] }
               sort { ($qshape{$b}->[0] <=> $qshape{$a}->[0]) or ($a cmp $b) }
               keys %qshape;
    
    # limit the data returned
    @data = splice @data, 0, $self->{max_subdomains};

    return \@data;
}

sub _fetch_qshape {
    my ($self, $server) = @_;
    my $queue = $self->{queue} ? lc($self->{queue}) : 'default';
    # hit http://$server:28080/services/quarantine/qshape/$queue
    my $url = "http://$server:28080/services/quarantine/qshape/$queue";

    my $ua = LWP::UserAgent->new;
    my $resp = $ua->get($url);
    unless ($resp->is_success) {
        die "Could not fetch $url: " . $resp->status_line;
    }
    return $resp->content;
}

1;
