package Sophos::Reports::ReportMailer;
use strict;
use warnings;
use Data::Dumper;
use Date::Format qw(time2str);
use PureMessage::Config qw(%config);
use PureMessage::Queue qw(queue_mail);
use PureMessage::Util::Email qw(is_valid_email);
use PureMessage::Reports::Util qw(get_hostname);
use PureMessage::Reports::DB qw(get_dbh);
use MIME::Entity;
use PureMessage::Reports::Custom qw(get_report_data);
use base 'Exporter';
our @EXPORT_OK = qw(report_list usage process_options process_reports
		    mail_reports);

our $self;
our $opts;
my $errors;

# refactor this if there's ever an "index" of reports
my @report_list = qw(
    CategoriesByGroup
    MessageCategories
    MessagesBlocked
    MtaRejected
    PolicyMarkHits
    QueueStatus
    TopOtherRelays
    TopRelays
    TopSpamRecipients
    TopSpamRelays
    TopSpamSenders
    TopVirusRelays
    TopVirusSenders
    TopVirusTypes
);

my %valid_reports = map { $_ => 1 } @report_list;

my ($custom_reports, %custom_titles);

sub _generate_time_span {
    my $sdate = shift;
    my $edate = shift;

    $sdate =~ /^(\d{4})-(\d{2})-(\d{2})$/;
    my $s_epoch =  Time::Local::timelocal(0, 0, 0, $3, $2 - 1, $1 - 1900);

    $edate =~ /^(\d{4})-(\d{2})-(\d{2})$/;
    my $e_epoch =  Time::Local::timelocal(0, 0, 0, $3, $2 - 1, $1 - 1900);

    return ($e_epoch - $s_epoch) / 60;
}

sub process_options {
    $opts = shift;
    if ( $opts->{help} ) {
        die usage();
    }

    # get custom reports
    $custom_reports = get_report_data();
    foreach (keys %$custom_reports) {
        $custom_titles{$custom_reports->{$_}{'title'}} = $_;
    }

    if ( $opts->{list} ) {
        my $list = join("\r\n", @report_list);

        # list custom reports too only if available
        my $list_c = '';
        if (scalar keys %custom_titles) {
            $list_c = join("\r\n", sort keys %custom_titles);
        }

        $list .= "\n\n<Custom Reports>:\n" . $list_c if $list_c;
        die ($list, "\n");
    }

    if ( $opts->{'time-frame'} ) {
       die ("Invalid value for option time-frame\n" . usage())
           unless $opts->{'time-frame'} =~ /(\d+)([hdw])/;
       my %time_frame_map = ( d => "days", w => "weeks", h => "hours" );
       $opts->{span} = "-$1 " . $time_frame_map{$2};
       delete $opts->{'time-frame'};
    }
    else {
        if ( $opts->{start} ) {
            die ("Invalid value for start date\n" . usage())
                unless ( verify_date($opts->{start}) );
           die ("Start date specified without end date\n" . usage())
               unless $opts->{end};
           $opts->{date_range_start} = delete $opts->{start};
        }
        elsif (!$opts->{start} && !$opts->{'time-frame'} ) {
                die usage();
        }

        if ( $opts->{end} ) {
           die ("End date specified without start date\n" . usage())
               unless $opts->{date_range_start};
           unless ( verify_date($opts->{end}) ) {
               die "Invalid value for option end date\n" . usage();
           }
           die ("End date must be later than start date\n" . usage())
               unless ($opts->{end} gt $opts->{date_range_start});
           $opts->{date_range_end} = delete $opts->{end};
        }

        my $span = _generate_time_span($opts->{date_range_start},
                                      $opts->{date_range_end});

        $opts->{date} = delete $opts->{date_range_start};
        $opts->{span} = "$span minutes";
    }

    if ( $opts->{direction} ) {
    	my %direction_map = ( inbound => 1, outbound => 1, both => 1 );
    	die ("Invalid value for option direction\n" . usage()) 
    	    unless exists $direction_map{$opts->{direction}};
    }

    if ( $opts->{custom} ) {
    	die ("Custom report specified without any report title\n" . usage()) 
	    unless scalar @{$opts->{custom}};

    	my @mkey;
    	for my $c (@{$opts->{custom}}) {
    	    die ("Invalid custom report title \"$c\"\n" . usage()) 
         	unless exists $custom_titles{$c};
            push @mkey, $custom_titles{$c};
    	}

    	$opts->{custom} = \@mkey;
    }

    if ( $opts->{report} ) {
    	for my $report (@{$opts->{report}}) {
    	    die ("Invalid report $report\n" . usage()) 
         		unless exists $valid_reports{$report};
    	}
    }
    elsif (!$opts->{custom}){
        $opts->{report} = ['MessageCategories'];
    }
    else {
        $opts->{report} = [];
    }

    unless ( $opts->{mailto} && is_valid_email($opts->{mailto}) ) {
        $opts->{mailto} ||= '';
        die ("Invalid 'mailto' address: '$opts->{mailto}'\n" . usage());
    }
  
    if ( $opts->{mailfrom} && !is_valid_email($opts->{mailfrom}) ) {
        die ("Invalid 'mailfrom' address: '$opts->{mailfrom}'\n" . usage());
    }
    $opts->{mailfrom} ||= $config{admin}; 

    return 1;
}

sub process_reports {
    my @reports;

    if ($opts->{custom}) {
        my @customs = @{$opts->{custom}};
        delete $opts->{custom};
        my $pkg = "Sophos::Reports::Report::CustomMarkHits";
        eval "require $pkg";
        if($@) {
            die ("Could not load report packge: '$pkg': $@");
        }
        foreach my $mkey (@customs)
        {
            my $report = "$pkg"->new(mkey => $mkey, %$opts);

            my $info = {
                'title' => $custom_reports->{$mkey}{title},
                'csv'   => $report->csv(),
            };

            push( @reports, $info );
        }
    }

    foreach my $rid (@{$opts->{report}})
    {
        my $pkg = "Sophos::Reports::Report::${rid}";
        eval "require $pkg";
        if($@) {
            die ("Could not load report packge: '$pkg': $@");
        }

        my $report = "$pkg"->new(%$opts);

        my $info = {
            'title' => $rid,
            'csv'   => $report->csv(),
        };

        push( @reports, $info );
    }

    mail_reports(\@reports);
}

sub mail_reports {
    my $reports = shift;

    my $dbh = get_dbh();
    my $hostname = get_hostname($dbh);

    my %hdrs = (
        'To'      => $opts->{mailto},
        'From'    => $opts->{mailfrom},
        'Subject' => $opts->{subject} || "pmx-reports-mailer-v2 $hostname",
    );
        
    my $mimemsg = MIME::Entity->build( %hdrs,
                        'Type'  => 'text/plain',
                        'Data'  => _mk_text_body($reports, $hostname),
                  );

    ###########################################################################
    # Add any attachments that are necessary; any reports that had a
    # "Content-ID" generated so that they could be dropped in inline.
    ###########################################################################
    my $date = time2str('%Y-%m-%d', time);
    foreach my $rpt (@$reports)
    {
            my $title = $rpt->{'title'};
            $mimemsg->attach( 'Id'          => $title,
                              'Data'        => $rpt->{'csv'},
                              'Type'        => 'text/comma-separated-values', 
                              'Disposition' => 'attachment',
                              'Description' => "Report-$title-$date.csv",
                              'Filename'    => "Report-$title-$date.csv",
                              );
    }
    $mimemsg->print() if ($opts->{'debug'});

    ###########################################################################
    # Send the message along on its way.
    ###########################################################################
    my $env_from = _get_raw_addr( $opts->{'mailfrom'} );
    my $env_to   = _get_raw_addr( $opts->{'mailto'} );

    queue_mail( from    => $env_from, 
                to      => $env_to, 
                message => $mimemsg->stringify(),
              );
}

###############################################################################
# Returns the first raw e-mail address found in the given address(es).
###############################################################################
sub _get_raw_addr {
    my $addr = shift;
    $addr =~ s/.*<(.*)>.*/$1/;
    return $addr;
}

sub verify_date {
    my $date = shift;
    if ($date =~ /^(\d{4})-(\d{2})-(\d{2})$/) {
       return 1;
    }
    return 0;
}

sub _mk_text_body
{
    my $reports  = shift;
    my $hostname = shift;

    my $body = "\nReports for $hostname:\n";
    for (@$reports) {
        $body .= "\t" . $_->{'title'} . "\n";
    }
    $body .= "\n" . "-"x70 . "\n";
    $body .= "Report options:\n";
    foreach my $k (keys %$opts) {
        next if ($k =~ 'report');
        # Remove the prepended hyphen on the 'span' option
        $opts->{$k} =~ s/-// if($k =~ 'span');
        $body .= "$k: " . $opts->{$k} . "\n";
    }
    $body .= "\n\n";
    return $body;
}


sub usage { <<USAGE }
Usage:
    $self  --report --custom --direction --subject --mailto

  Options:
    -l, --list
            Prints a list of available reports that can be emailed or run from 
            the command line. Use the syntax displayed in this list when 
            specifying a report with the --report or --custom option.

    -m, --mailto
            The email address the report(s) will be mailed to. You must specify 
            at least one recipient. 
    
    -r, --report MessageCategories
            The report to run and mail. A list of all report names can be generated 
            with the --list option. You can specify multiple reports if you want
            to mail more than one. If no report is specified, the default 
            MessageCategories is used. For example, to email the ``Top Spam 
            Senders'' report to me\@example.com, enter the following at the 
            command line:

        pmx-reports-mailer-v2 --report TopSpamSenders --mailto me\@example.com

    -c, --custom "Custom Report Foo" 
            The custom report to run and mail. A list of all custom report
            titles can be generated with the --list option. You can specify multiple 
            titles if you want to mail more than one. For example, to email the
            ``Custom Report Foo'' report to me\@example.com, enter the following at 
            the command line:

        pmx-reports-mailer-v2 --custom "Custom Report Foo" --mailto me\@example.com

    -d, --direction both
            Specify the direction (inbound,outbound,or both) for the report data.
            By default, the data will be given for both directions.

    -s, --start "YYYY-MM-DD"
    -e, --end "YYYY-MM-DD"
            Specify the date range for the report. A start date and end date 
            must be specified together. 
            You can also use the --time-frame option (described below) as an 
            alternative to start and/or end. The command must include one of 
            these two methods.

    -t, --time-frame 24h
            The time frame that is reported on. Values can be specified as any 
            number of hours or days (e.g. 24h, 3d). The default is 24h. As an 
            alternative method, use the --start and --end options described 
            above. The command must include one of these two methods.

    -j, --subject
            The subject of the report. For example:

        Subject: Top Spam Senders

            By default, the subject is pmx-reports-mailer-v2 \$hostname, where 
            \$hostname is the name of the server from which the repo

    -h, --help
            Print this help text.

USAGE
1;
